/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SCOREDSEQCOLLECTIONBWT_CPP
#define SCOREDSEQCOLLECTIONBWT_CPP

#include "ScoredSeqCollectionBwt.h"

#include "AssemblyException.h"
#include "AlignmentNull.h"
#include "AlignmentUngapped.h"
#include "ScoredSeqFlip.h"
#include <iostream>
#include <omp.h>
#include <algorithm>
using namespace::std;


ScoredSeqCollectionBwt::ScoredSeqCollectionBwt(){}

ScoredSeqCollectionBwt::ScoredSeqCollectionBwt(set<ScoredSeq*>* inputSeqs, float fractId) : 
  _asi(0),
  _gapMode(ungapped),
  _fractId(fractId),
  _minOverlap(0),
  _penalizeEdgeGaps(true),
  _usingMaxOverlap(false),
  _maxOverlap(-1) {
  constructorHelper(inputSeqs);
}
ScoredSeqCollectionBwt::ScoredSeqCollectionBwt(set<ScoredSeq*>* inputSeqs, float fractId, long minOverlap) : 
  _asi(0),
  _gapMode(ungapped),
  _fractId(fractId),
  _minOverlap(minOverlap),
  _penalizeEdgeGaps(false),
  _usingMaxOverlap(false),
  _maxOverlap(-1) {
  constructorHelper(inputSeqs);
}
ScoredSeqCollectionBwt::ScoredSeqCollectionBwt(set<ScoredSeq*>* inputSeqs, float fractId, long minOverlap, long maxOverlap) : 
  _asi(0),
  _gapMode(ungapped),
  _fractId(fractId),
  _minOverlap(minOverlap),
  _penalizeEdgeGaps(false),
  _usingMaxOverlap(true),
  _maxOverlap(maxOverlap) {
  constructorHelper(inputSeqs);
}
ScoredSeqCollectionBwt::ScoredSeqCollectionBwt(set<ScoredSeq*>* inputSeqs, DynamicProgrammingAligner * asi) :
  _gapMode(gapped),
  _fractId(asi->getFractId()),
  _minOverlap(asi->getMinOverlap()),
  _penalizeEdgeGaps(false),
  _usingMaxOverlap(false),
  _maxOverlap(-1) {
  _asi = new DynamicProgrammingAligner(asi);
  constructorHelper(inputSeqs);
}
ScoredSeqCollectionBwt::ScoredSeqCollectionBwt(set<ScoredSeq*>* inputSeqs, DynamicProgrammingAligner * asi, long maxOverlap) :
  _gapMode(gapped),
  _fractId(asi->getFractId()),
  _minOverlap(asi->getMinOverlap()),
  _penalizeEdgeGaps(false),
  _usingMaxOverlap(true),
  _maxOverlap(maxOverlap) {
  _asi = new DynamicProgrammingAligner(asi);
  constructorHelper(inputSeqs);
}

void ScoredSeqCollectionBwt::constructorHelper(set<ScoredSeq*>* inputSeqs){
  // by default, edge scaling will happen.  it can be disabled to save
  // time during read mapping.
  _edgeScalingEnabled = true;

  // minOverlap cannot be zero for this class; sub-string words must be defined
  // and statistically scaled for this overlap minimum, so a hard bottom line
  // needs to be defined here
  long minOverlapLimit = 4;
  if (_minOverlap < minOverlapLimit){ _minOverlap = minOverlapLimit; }

  // fractId must be at least that expected for random sequence matches
  if (_fractId < 0.25){ _fractId = 0.25; }

  // determine the length of the shortest sequence and whether or not it is
  // smaller than _minOverlap
  long minSeqLength = 0;
  set<ScoredSeq*>::iterator seqIt = inputSeqs->begin();
  if (seqIt != inputSeqs->end()){
    minSeqLength = (*seqIt)->size();
    ++seqIt;
  }
  while (seqIt != inputSeqs->end()){
    if ((*seqIt)->size() < minSeqLength){ minSeqLength = (*seqIt)->size(); }
    ++seqIt;
  }
  if (minSeqLength < _minOverlap){ _minFragmentSize = _minOverlap; }
  else { _minFragmentSize = long(float(minSeqLength) * _fractId); }

  // some data that will be useful to have pre-computed
  _maxFractMis = 1.0 - _fractId;
  long matchScore = 1;
  long misScore = -1;
  _scoreAsm = new AlignmentScoreMatrix(matchScore,misScore,misScore,misScore);
  _matchCountAsm = new AlignmentScoreMatrix(1,0,0,0);
  _misCountAsm = new AlignmentScoreMatrix(0,1,1,1); // gaps count as mismatches (terminal gaps will exist)
  _allSeqs.insert( inputSeqs->begin(), inputSeqs->end() );

  // figure out the windows and make the bins
  long seqCount = 0;
  long seqLenSum = 0;
  for (set<ScoredSeq*>::iterator it = inputSeqs->begin(); it != inputSeqs->end(); ++it){
    seqCount++;
    seqLenSum += (*it)->size() + 1; // +1 because of the separator/end char
  }
  if (seqCount==0){
    _binSize = 0; // this value is only possible in this case
    _maxBin = 0;
  } else {
    _binSize = seqLenSum / seqCount + 1; // +1 to round up
    _maxBin = seqLenSum / _binSize;
  }
  _binToSeqs = new set<ScoredSeqLocus*>*[_maxBin+1];

  
  for (long b = 0; b <= _maxBin; ++b){
    _binToSeqs[b] = new set<ScoredSeqLocus*>;
  }

  // convert the full text to a numeric array, filling in the bins
  if (seqLenSum==0){ seqLenSum = 1; } // at least an end char will be added
  _textSize = seqLenSum;

  // a local array of text length
  long* textAsNums = new long[seqLenSum];

  long currentIndex = 0;
  _numContigLoci = inputSeqs->size();
  if (_numContigLoci > 0){
    _orderedContigLoci = new ScoredSeqLocus*[ _numContigLoci ];
  }

  long currentOffset = 0;
  for (set<ScoredSeq*>::iterator it = _allSeqs.begin(); it != _allSeqs.end(); ++it){
    // add the sequence, preceeded by a separator if necessary
    if ( it != _allSeqs.begin() ){
      char* separator = new char[2];
      separator[0] = '&';
      separator[1] = '\0';
      burrowsWheelerUtilities::makeNumString( separator, long(1), currentOffset, textAsNums );
      delete [] separator;
      currentOffset++;
    }
    ScoredSeq* seq = (*it);
    char* seqString = seq->getSeq('+');
    burrowsWheelerUtilities::makeNumString( seqString, seq->size(), currentOffset, textAsNums );
    delete [] seqString;

    // fill the overlapping bins with this ScoredSeq in Locus form
    long firstBin = currentOffset / _binSize;
    long lastBinP1 = ( currentOffset + seq->size() ) / _binSize + 1;
    ScoredSeqLocus* seqAsLocus = new ScoredSeqLocus(seq, currentOffset, currentOffset + seq->size() - 1, currentIndex);
    for (long bin = firstBin; bin < lastBinP1; ++bin){ _binToSeqs[bin]->insert( seqAsLocus ); }

    _orderedContigLoci[currentIndex] = seqAsLocus;
    currentIndex++;

    currentOffset += seq->size();
  }

  // end the sequence with an end char (even if there were no sequence entries)
  char* endChar = new char[2];
  endChar[0] = '#';
  endChar[1] = '\0';
  burrowsWheelerUtilities::makeNumString( endChar, long(1), currentOffset, textAsNums );
  delete [] endChar;

  // make the burrows-wheeler transform
  _sortedSuffixes = new long[seqLenSum];
  long alphaSize = burrowsWheelerUtilities::getAlphabetSize();
  burrowsWheelerUtilities::sortSuffixes( textAsNums, _sortedSuffixes, seqLenSum, alphaSize );
  _bwTransform = new long[seqLenSum];
  burrowsWheelerUtilities::bwTransform(textAsNums, _sortedSuffixes, seqLenSum, _bwTransform);

  delete [] textAsNums;

  // make the accessory data structures for alignment
  _occCounts = new long[seqLenSum];
  _tableC = burrowsWheelerUtilities::fillBwtHelperArrays(_bwTransform, seqLenSum, alphaSize, _occCounts);
  //OK();
}



void ScoredSeqCollectionBwt::OK(){
  if ( _fractId > 1 ){
    throw AssemblyException::ArgError("fraction ID is a fraction; cannot be greater than 1");
  }
  if (_allSeqs.size()==0){
    if (_textSize != 1){ throw AssemblyException::LogicError("BWT text len is 1 if there are no seqs"); }
  } else {
    long seqLenSum = 0;
    for (set<ScoredSeq*>::iterator it = _allSeqs.begin(); it != _allSeqs.end(); ++it){
      seqLenSum += (*it)->size() + 1; // +1 because of the separator/end char
    }
    if (_textSize != seqLenSum){ throw AssemblyException::LogicError("BWT text len is wrong"); }
  }
}



ScoredSeqCollectionBwt::~ScoredSeqCollectionBwt(){
  delete _asi;
  delete _matchCountAsm;
  delete _misCountAsm;
  delete _scoreAsm;
  delete [] _sortedSuffixes;
  delete [] _bwTransform;
  delete [] _occCounts;
  delete [] _tableC;
  for (long n = 0; n < _numContigLoci; ++n){ delete _orderedContigLoci[n]; }
  if (_numContigLoci > 0){ delete [] _orderedContigLoci; }

  for (long b = 0; b <= _maxBin; ++b){ delete _binToSeqs[b]; }
  delete [] _binToSeqs;
}




ScoredSeqCollectionBwt * ScoredSeqCollectionBwt::copy(){
  // uses a private constructor to efficiently copy stuff
  throw AssemblyException::ImplementationError("not implemented");
}

float ScoredSeqCollectionBwt::getFractId(){ return _fractId; }



// DEFAULT: edge scaling is enabled
void ScoredSeqCollectionBwt::enableEdgeScaling(){ _edgeScalingEnabled = true; }
void ScoredSeqCollectionBwt::disableEdgeScaling(){ _edgeScalingEnabled = false; }



bool ScoredSeqCollectionBwt::contains(ScoredSeq* s){
  //OK();
  return ( _allSeqs.find(s) != _allSeqs.end() );
}



void ScoredSeqCollectionBwt::getSeqs( set<ScoredSeq*>* allSeqs ){
  //OK();
  allSeqs->insert( _allSeqs.begin(), _allSeqs.end() );
}





void ScoredSeqCollectionBwt::getMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* matchTest){
  set<Alignment*> tempMatches;
  getMatchesPrivate(&tempMatches, seq, matchTest, _minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
}

void ScoredSeqCollectionBwt::getMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* matchTest){
  set<Alignment*> tempMatches;
  getMatchesPrivate(&tempMatches, seq, sense, matchTest, _minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
}


void ScoredSeqCollectionBwt::getMatches(set<Alignment*>* matches, ScoredSeq* seq){
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  getMatchesPrivate(&tempMatches, seq, seqTest, _minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
}

void ScoredSeqCollectionBwt::getMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense){
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  getMatchesPrivate(&tempMatches, seq, sense, seqTest, _minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
}


// min overlap is specified
void ScoredSeqCollectionBwt::getMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* matchTest,
					long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<Alignment*> tempMatches;
  getMatchesPrivate(&tempMatches, seq, matchTest, minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
}
void ScoredSeqCollectionBwt::getMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* matchTest,
					long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<Alignment*> tempMatches;
  getMatchesPrivate(&tempMatches, seq, sense, matchTest, minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
}
void ScoredSeqCollectionBwt::getMatches(set<Alignment*>* matches, ScoredSeq* seq,
					long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  getMatchesPrivate(&tempMatches, seq, seqTest, minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
}
void ScoredSeqCollectionBwt::getMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense,
					long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  getMatchesPrivate(&tempMatches, seq, sense, seqTest, minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
}


// ASSUMES: matches is empty; otherwise, it will be modified
// WILL RETURN MULTIPLE DISTINCT ALIGNMENTS TO A SINGLE TARGET
void ScoredSeqCollectionBwt::getMatchesPrivate(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* matchTest, long minOverlap){
  //OK();
  getMatchesPrivate(matches, seq, '+', matchTest, minOverlap);
  getMatchesPrivate(matches, seq, '-', matchTest, minOverlap);
  //OK();
}


// REQUIRES: all members of "matches" (if any) have the same score
void ScoredSeqCollectionBwt::getMatchesPrivate(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* matchTest, long minOverlap){
  bufferSeqs();

  long querySize = seq->size();
  AlignmentMaker * alMaker;
  if (_penalizeEdgeGaps){
    long maxMismatch = long(float(querySize) * _maxFractMis); // keep constant; this isn't best matches
    alMaker = new AlMakerPenalizeEdgeA(_scoreAsm, -10, _maxFractMis); // value in alMaker
  } else {
    alMaker = new AlMakerNoEdge(_scoreAsm, -10, _maxFractMis); // value in alMaker
  }

  // get the numeric representation of the query
  ScoredSeq* seqFlip = ScoredSeqFlip::getFlip(seq, sense);

  // for each matched contig, keeps track of the num hits with any
  // given offset
  OffsetTracker** contigToOffsets = getOffsets(seq,sense,minOverlap,matchTest,semiGlobal);

  // make alignments
  long hitIndex = 0;
  while (contigToOffsets[hitIndex] != NULL){
    OffsetTracker* tracker = contigToOffsets[hitIndex];
    hitIndex++;
    ScoredSeq* contig = tracker->targetContig();
    long* allOffsets = tracker->getOffsets();
    long numOffsets = tracker->numOffsets();
    delete tracker;
    set<long> adequateOffsets;
    for (long offIndex = 0; offIndex < numOffsets; ++offIndex){
      // determine how long the alignment will be??  i will need to do that to make sure
      // that only legit alignments are kept.
      long offset = allOffsets[offIndex];
      long seqStart;
      if (offset < 0){ seqStart = 0; }
      else { seqStart = offset; }
      long seqEnd;
      if (offset + contig->size() >= seq->size()){ seqEnd = seq->size(); }
      else { seqEnd = offset + contig->size(); }
      long alLength = seqEnd - seqStart;
      if (alLength < 0){ throw AssemblyException::LogicError("SSCBwt::getMatches, alLength should be > 0"); }
      if ( (_penalizeEdgeGaps or alLength >= minOverlap) and ( (! _usingMaxOverlap) or alLength <= _maxOverlap) ){
	adequateOffsets.insert(offset);
      }
    }
    delete [] allOffsets;

    // this will be different for gapped versus ungapped alignments
    if (_gapMode == ungapped){ // UNGAPPED ALIGNMENT
      for (set<long>::iterator offsetIt = adequateOffsets.begin(); offsetIt != adequateOffsets.end(); ++offsetIt){
	long offset = *offsetIt;

	Alignment* alignCand = alMaker->makeAlignment(seqFlip, contig, '+', offset);

	if ( alignCand->isNull() ){ delete alignCand; }
	else {
	  if (sense == '+'){
	    alignCand->seqReplace(seq, alignCand->seqB());
	    matches->insert(alignCand);
	  } else {
	    matches->insert( alignCand->copyRcSeqA(seq) );
	    //matches->insert( reverseAlignmentSeqA(alignCand, seq) );
	    delete alignCand;
	  }
	}
      }
    } else { // GAPPED ALIGNMENT
      if (adequateOffsets.size() > 0){
	Alignment* alignCand = _asi->align(seqFlip,contig,'+',&adequateOffsets);
	if ( alignCand->isNull() ){ delete alignCand; }
	else {
	  if (sense == '+'){
	    alignCand->seqReplace(seq, alignCand->seqB());
	    matches->insert(alignCand);
	  } else {
	    matches->insert( alignCand->copyRcSeqA(seq) );
	    //matches->insert( reverseAlignmentSeqA(alignCand, seq) );
	    delete alignCand;
	  }
	}
      }
    }
  }
  delete [] contigToOffsets;

  delete alMaker;
  delete seqFlip;
  //if (currentAsi != asi){ delete currentAsi; }
}



void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* seqTest, AlignmentType alType){
  long localMin;
  if (alType == semiGlobal){ localMin = _minOverlap; }
  else if (alType == fullAlignment){ localMin = _minFragmentSize; }
  else { throw AssemblyException::ArgError("SSCBwt:getBestMatch got bad alType"); }
  set<Alignment*> tempMatches;
  getBestMatchesPrivate(&tempMatches, seq, seqTest, localMin, alType);
  matches->insert( tempMatches.begin(), tempMatches.end() );
}

void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* seqTest, AlignmentType alType){
  long localMin;
  if (alType == semiGlobal){ localMin = _minOverlap; }
  else if (alType == fullAlignment){ localMin = _minFragmentSize; }
  else { throw AssemblyException::ArgError("SSCBwt:getBestMatch got bad alType"); }
  set<Alignment*> tempMatches;
  getBestMatchesPrivate(&tempMatches, seq, sense, seqTest, localMin, alType);
  matches->insert( tempMatches.begin(), tempMatches.end() );
}


void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, AlignmentType alType){
  long localMin;
  if (alType == semiGlobal){ localMin = _minOverlap; }
  else if (alType == fullAlignment){ localMin = _minFragmentSize; }
  else { throw AssemblyException::ArgError("SSCBwt:getBestMatch got bad alType"); }
  set<ScoredSeq*> okTargets; // a dummy container just to fulfill the specs
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  getBestMatchesPrivate(&tempMatches, seq, seqTest, localMin, alType);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
}

void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, AlignmentType alType){
  long localMin;
  if (alType == semiGlobal){ localMin = _minOverlap; }
  else if (alType == fullAlignment){ localMin = _minFragmentSize; }
  else { throw AssemblyException::ArgError("SSCBwt:getBestMatch got bad alType"); }
  set<ScoredSeq*> okTargets; // a dummy container just to fulfill the specs
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  getBestMatchesPrivate(&tempMatches, seq, sense, seqTest, localMin, alType);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
}



// min overlap is specified
void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* seqTest,
					    long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<Alignment*> tempMatches;
  getBestMatchesPrivate(&tempMatches, seq, seqTest, minOverlap, semiGlobal);
  matches->insert( tempMatches.begin(), tempMatches.end() );
}
void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* seqTest,
					    long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<Alignment*> tempMatches;
  getBestMatchesPrivate(&tempMatches, seq, sense, seqTest, minOverlap, semiGlobal);
  matches->insert( tempMatches.begin(), tempMatches.end() );
}
void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq,
					    long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<ScoredSeq*> okTargets; // a dummy container just to fulfill the specs
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  getBestMatchesPrivate(&tempMatches, seq, seqTest, minOverlap, semiGlobal);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
}
void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense,
					    long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<ScoredSeq*> okTargets; // a dummy container just to fulfill the specs
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  getBestMatchesPrivate(&tempMatches, seq, sense, seqTest, minOverlap, semiGlobal);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
}



// THE BEGINNING OF HASMATCH

bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, MatchSeqTest* seqTest, AlignmentType alType){
  long localMin;
  if (alType == semiGlobal){ localMin = _minOverlap; }
  else if (alType == fullAlignment){ localMin = _minFragmentSize; }
  else { throw AssemblyException::ArgError("SSCBwt:hasMatch got bad alType"); }
  return hasMatchPrivate(seq, seqTest, localMin, alType);
}

bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, char sense, MatchSeqTest* seqTest, AlignmentType alType){
  long localMin;
  if (alType == semiGlobal){ localMin = _minOverlap; }
  else if (alType == fullAlignment){ localMin = _minFragmentSize; }
  else { throw AssemblyException::ArgError("SSCBwt:hasMatch got bad alType"); }
  return hasMatchPrivate(seq, sense, seqTest, localMin, alType);
}


bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, AlignmentType alType){
  long localMin;
  if (alType == semiGlobal){ localMin = _minOverlap; }
  else if (alType == fullAlignment){ localMin = _minFragmentSize; }
  else { throw AssemblyException::ArgError("SSCBwt:hasMatch got bad alType"); }
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  bool hasMatch = hasMatchPrivate(seq, seqTest, localMin, alType);
  delete seqTest;
  return hasMatch;
}

bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, char sense, AlignmentType alType){
  long localMin;
  if (alType == semiGlobal){ localMin = _minOverlap; }
  else if (alType == fullAlignment){ localMin = _minFragmentSize; }
  else { throw AssemblyException::ArgError("SSCBwt:hasMatch got bad alType"); }
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  bool hasMatch = hasMatchPrivate(seq, sense, seqTest, localMin, alType);
  delete seqTest;
  return hasMatch;
}




// min overlap is specified
bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, MatchSeqTest* seqTest, long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  return hasMatchPrivate(seq, seqTest, minOverlap, semiGlobal);
}
bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, char sense, MatchSeqTest* seqTest, long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  return hasMatchPrivate(seq, sense, seqTest, minOverlap, semiGlobal);
}
bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  bool hasMatch = hasMatchPrivate(seq, seqTest, minOverlap, semiGlobal);
  delete seqTest;
  return hasMatch;
}
bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, char sense, long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  bool hasMatch = hasMatchPrivate(seq, sense, seqTest, minOverlap, semiGlobal);
  delete seqTest;
  return hasMatch;
}



// THE END OF HASMATCH



// ASSUMES: matches is empty; otherwise, it will be modified
// WILL RETURN MULTIPLE DISTINCT ALIGNMENTS TO A SINGLE TARGET
void ScoredSeqCollectionBwt::getBestMatchesPrivate(set<Alignment*>* matches, ScoredSeq* seq,
						   MatchSeqTest* seqTest, long minOverlap, AlignmentType alType){
  //OK();
  // for each matched contig, keeps track of the num hits with any
  // given offset
  OffsetTracker** ctoPlus = getOffsets(seq,'+',minOverlap,seqTest,alType);
  OffsetTracker** ctoMinus = getOffsets(seq,'-',minOverlap,seqTest,alType);

  getBestMatchesPrivateUngapped(ctoPlus, matches, seq, '+', seqTest, minOverlap);
  getBestMatchesPrivateUngapped(ctoMinus, matches, seq, '-', seqTest, minOverlap);
  if (_gapMode == gapped){
    long bestScoreSoFar = 0;
    if (matches->size() > 0){ bestScoreSoFar = (*(matches->begin()))->score(_scoreAsm,_penalizeEdgeGaps); }
    for (set<Alignment*>::iterator it = matches->begin(); it != matches->end(); ++it){ delete *it; }
    matches->clear();
    getBestMatchesPrivateGapped(ctoPlus, matches, seq, '+', seqTest, minOverlap, bestScoreSoFar);
    getBestMatchesPrivateGapped(ctoMinus, matches, seq, '-', seqTest, minOverlap, bestScoreSoFar);
  }

  // delete the offset tracker arrays and their contents
  long hitIndex = 0;
  while (ctoPlus[hitIndex] != NULL){
    delete ctoPlus[hitIndex];
    hitIndex++;
  }
  delete [] ctoPlus;
  hitIndex = 0;
  while (ctoMinus[hitIndex] != NULL){
    delete ctoMinus[hitIndex];
    hitIndex++;
  }
  delete [] ctoMinus;
  //OK();
}
// REQUIRES: all members of "matches" (if any) have the same score
void ScoredSeqCollectionBwt::getBestMatchesPrivate(set<Alignment*>* matches, ScoredSeq* seq, char sense,
						   MatchSeqTest* seqTest, long minOverlap, AlignmentType alType){
  // for each matched contig, keeps track of the num hits with any
  // given offset
  OffsetTracker** contigToOffsets = getOffsets(seq,sense,minOverlap,seqTest,alType);
  getBestMatchesPrivateUngapped(contigToOffsets, matches, seq, sense, seqTest, minOverlap);
  if (_gapMode == gapped){
    long bestScoreSoFar = 0;
    if (matches->size() > 0){ bestScoreSoFar = (*(matches->begin()))->score(_scoreAsm,_penalizeEdgeGaps); }
    for (set<Alignment*>::iterator it = matches->begin(); it != matches->end(); ++it){ delete *it; }
    matches->clear();
    getBestMatchesPrivateGapped(contigToOffsets, matches, seq, sense, seqTest, minOverlap, bestScoreSoFar);
  }
  // delete array and contents
  long hitIndex = 0;
  while (contigToOffsets[hitIndex] != NULL){
    delete contigToOffsets[hitIndex];
    hitIndex++;
  }
  delete [] contigToOffsets;
  //OK();
}




// ASSUMES: matches is empty; otherwise, it will be modified
// WILL RETURN MULTIPLE DISTINCT ALIGNMENTS TO A SINGLE TARGET
bool ScoredSeqCollectionBwt::hasMatchPrivate(ScoredSeq* seq, MatchSeqTest* seqTest, long minOverlap, AlignmentType alType){

  set<Alignment*> matches;
  // for each matched contig, keeps track of the num hits with any
  // given offset
  OffsetTracker** ctoPlus = getOffsets(seq,'+',minOverlap,seqTest,alType);
  OffsetTracker** ctoMinus = getOffsets(seq,'-',minOverlap,seqTest,alType);

  getBestMatchesPrivateUngapped(ctoPlus, &matches, seq, '+', seqTest, minOverlap, true);
  if (matches.size()==0){ getBestMatchesPrivateUngapped(ctoMinus, &matches, seq, '-', seqTest, minOverlap, true); }
  if (matches.size()==0 and _gapMode == gapped){
    long bestScoreSoFar = 0;
    getBestMatchesPrivateGapped(ctoPlus, &matches, seq, '+', seqTest, minOverlap, bestScoreSoFar, true);
    if (matches.size()==0){ getBestMatchesPrivateGapped(ctoMinus, &matches, seq, '-', seqTest, minOverlap, bestScoreSoFar, true); }
  }

  // delete the offset tracker arrays and their contents
  long hitIndex = 0;
  while (ctoPlus[hitIndex] != NULL){
    delete ctoPlus[hitIndex];
    hitIndex++;
  }
  delete [] ctoPlus;
  hitIndex = 0;
  while (ctoMinus[hitIndex] != NULL){
    delete ctoMinus[hitIndex];
    hitIndex++;
  }
  delete [] ctoMinus;
  bool hasMatch = matches.size() > 0;
  for (set<Alignment*>::iterator it = matches.begin(); it != matches.end(); ++it){ delete *it; }
  return hasMatch;
}
// REQUIRES: all members of "matches" (if any) have the same score
bool ScoredSeqCollectionBwt::hasMatchPrivate(ScoredSeq* seq, char sense, MatchSeqTest* seqTest, long minOverlap, AlignmentType alType){
  set<Alignment*> matches;
  // for each matched contig, keeps track of the num hits with any
  // given offset
  OffsetTracker** contigToOffsets = getOffsets(seq,sense,minOverlap,seqTest,alType);
  getBestMatchesPrivateUngapped(contigToOffsets, &matches, seq, sense, seqTest, minOverlap, true);
  if (matches.size()==0 and _gapMode == gapped){
    long bestScoreSoFar = 0;
    getBestMatchesPrivateGapped(contigToOffsets, &matches, seq, sense, seqTest, minOverlap, bestScoreSoFar, true);
  }
  // delete array and contents
  long hitIndex = 0;
  while (contigToOffsets[hitIndex] != NULL){
    delete contigToOffsets[hitIndex];
    hitIndex++;
  }
  delete [] contigToOffsets;
  bool hasMatch = matches.size() > 0;
  for (set<Alignment*>::iterator it = matches.begin(); it != matches.end(); ++it){ delete *it; }
  return hasMatch;
}



// THESE WILL REPLACE THE SINGLE METHOD BELOW
// REQUIRES: all members of "matches" (if any) have the same score
void ScoredSeqCollectionBwt::getBestMatchesPrivateUngapped(OffsetTracker** contigToOffsets,
							   set<Alignment*>* matches, ScoredSeq* seq, char sense,
							   MatchSeqTest* seqTest, long minOverlap, bool justFirstMatch){
  bufferSeqs();

  if (matches->size() > 0 and justFirstMatch){
    throw AssemblyException::CallingError("SSCBwt:getBestMatchesPrivateUngapped should not be called for justOneMatch if one was already found");
  }

  // must compete with prior matches (if any), but zero is the baseline
  long bestScoreSoFar = 0;
  if (matches->size() > 0){ bestScoreSoFar = (*(matches->begin()))->score(_scoreAsm,_penalizeEdgeGaps); }

  long querySize = seq->size();
  AlignmentMaker * alMaker;
  if (_penalizeEdgeGaps){
    long maxMismatch = long(float(querySize) * _maxFractMis); // keep constant; this isn't best matches
    alMaker = new AlMakerPenalizeEdgeA(_scoreAsm, -10, _maxFractMis); // value in alMaker
  } else {
    alMaker = new AlMakerNoEdge(_scoreAsm, -10, _maxFractMis); // value in alMaker
  }

  // get the numeric representation of the query
  ScoredSeq* seqFlip = ScoredSeqFlip::getFlip(seq, sense);

  // make alignments
  long hitIndex = 0;
  while (contigToOffsets[hitIndex] != NULL and ( (! justFirstMatch) or matches->size() == 0) ){
    OffsetTracker* contigIt = contigToOffsets[hitIndex];
    hitIndex++;
    // get the subset of offsets that are good enough for alignment according
    // to the global parameters
    ScoredSeq* contig = contigIt->targetContig();

    long* allOffsets = contigIt->getOffsets();
    long numOffsets = contigIt->numOffsets();
    long* adequateOffsets = new long[ numOffsets + 1 ];
    long numAdequateOffsets = 0;
    for (long allIndex = 0; allIndex < numOffsets; ++allIndex){
      // determine how long the alignment will be??  i will need to do that to make sure
      // that only legit alignments are kept.
      long offset = allOffsets[allIndex];
      long seqStart;
      if (offset < 0){ seqStart = 0; }
      else { seqStart = offset; }
      long seqEnd;
      if (offset + contig->size() >= seq->size()){ seqEnd = seq->size(); }
      else { seqEnd = offset + contig->size(); }
      long alLength = seqEnd - seqStart;
      if (alLength < 0){ throw AssemblyException::LogicError("SSCBwt::getBestMatches, alLength should be > 0"); }
      if ( (_penalizeEdgeGaps or alLength >= minOverlap) and ( (! _usingMaxOverlap) or alLength <= _maxOverlap) ){
	adequateOffsets[numAdequateOffsets] = offset;
	numAdequateOffsets++;
      }
    }
    delete [] allOffsets;

    for (long adequateIndex = 0; adequateIndex < numAdequateOffsets; ++adequateIndex){
      // this will speed things up in the just-one-match case by preventing alignments from being generated
      if ( (! justFirstMatch) or matches->size() == 0){
	long offset = adequateOffsets[adequateIndex];

	Alignment* alignCand = alMaker->makeAlignment(seqFlip, contig, '+', offset);
	long newScore;
	if (! alignCand->isNull()){ newScore = alMaker->getScore(alignCand); }
        if ( alignCand->isNull() or (matches->size() > 0 and newScore < bestScoreSoFar) ){
	  delete alignCand;
	} else {
          // replace the old alignments of the score is bigger; also, set the best score in
          // that case or if this is the first legit alignment
	  if (matches->size()==0 or newScore > bestScoreSoFar){
	    bestScoreSoFar = newScore;
	    for (set<Alignment*>::iterator oldIt = matches->begin(); oldIt != matches->end(); ++oldIt){ delete (*oldIt); }
	    matches->clear();
	    alMaker->setMinScore( newScore );
	  }
          // add the alignment to the set of matches
	  if (matches->size()==0 or newScore >= bestScoreSoFar){
	    if (sense == '+'){
	      alignCand->seqReplace(seq, alignCand->seqB());
	      matches->insert(alignCand);
	    } else {
	      matches->insert( alignCand->copyRcSeqA(seq) );
	      delete alignCand;
	    }
	  } else { delete alignCand; }
	}
      }
    }
    delete [] adequateOffsets;
  }

  delete alMaker;
  delete seqFlip;
}

// REQUIRES: all members of "matches" (if any) have the same score
void ScoredSeqCollectionBwt::getBestMatchesPrivateGapped(OffsetTracker** contigToOffsets,
							 set<Alignment*>* matches, ScoredSeq* seq, char sense,
							 MatchSeqTest* seqTest,
							 long minOverlap, long bestScoreSoFar, bool justFirstMatch){
  bufferSeqs();

  if (matches->size() > 0 and justFirstMatch){
    throw AssemblyException::CallingError("SSCBwt:getBestMatchesPrivateGapped should not be called for justOneMatch if one was already found");
  }

  // must compete with prior matches (if any), but zero is the baseline
  if (matches->size() > 0){ bestScoreSoFar = (*(matches->begin()))->score(_scoreAsm,_penalizeEdgeGaps); }

  long querySize = seq->size();
  AlignmentMaker * alMaker;
  if (_penalizeEdgeGaps){
    long maxMismatch = long(float(querySize) * _maxFractMis); // keep constant; this isn't best matches
    alMaker = new AlMakerPenalizeEdgeA(_scoreAsm, -10, _maxFractMis); // value in alMaker
  } else {
    alMaker = new AlMakerNoEdge(_scoreAsm, -10, _maxFractMis); // value in alMaker
  }

  // get the numeric representation of the query
  ScoredSeq* seqFlip = ScoredSeqFlip::getFlip(seq, sense);

  // make alignments
  long hitIndex = 0;
  while (contigToOffsets[hitIndex] != NULL and ( (! justFirstMatch) or matches->size() == 0) ){
    OffsetTracker* contigIt = contigToOffsets[hitIndex];
    hitIndex++;
    // get the subset of offsets that are good enough for alignment according
    // to the global parameters
    ScoredSeq* contig = contigIt->targetContig();

    long* allOffsets = contigIt->getOffsets();
    long numOffsets = contigIt->numOffsets();
    long* adequateOffsets = new long[ numOffsets + 1 ];
    long numAdequateOffsets = 0;
    for (long allIndex = 0; allIndex < numOffsets; ++allIndex){
      // determine how long the alignment will be??  i will need to do that to make sure
      // that only legit alignments are kept.
      long offset = allOffsets[allIndex];
      long seqStart;
      if (offset < 0){ seqStart = 0; }
      else { seqStart = offset; }
      long seqEnd;
      if (offset + contig->size() >= seq->size()){ seqEnd = seq->size(); }
      else { seqEnd = offset + contig->size(); }
      long alLength = seqEnd - seqStart;
      if (alLength < 0){ throw AssemblyException::LogicError("SSCBwt::getBestMatches, alLength should be > 0"); }
      if ( (_penalizeEdgeGaps or alLength >= minOverlap) and ( (! _usingMaxOverlap) or alLength <= _maxOverlap) ){
	adequateOffsets[numAdequateOffsets] = offset;
	numAdequateOffsets++;
      }
    }
    delete [] allOffsets;

    // set up an aligner that will dynamically set new minScores
    DynamicProgrammingAligner * localAsi = _asi->minScoreCopy(bestScoreSoFar);

    if (numAdequateOffsets > 0){
      Alignment* alignCand = localAsi->align(seqFlip,contig,'+',adequateOffsets,numAdequateOffsets);
      if ( alignCand->isNull() ){ delete alignCand; }
      else {
	long newBestScore = alignCand->score(_scoreAsm,_penalizeEdgeGaps);
	if (matches->size()==0 or newBestScore > bestScoreSoFar){
	  bestScoreSoFar = newBestScore;
	  DynamicProgrammingAligner * newLocalAsi = localAsi->minScoreCopy(bestScoreSoFar);
	  delete localAsi;
	  localAsi = newLocalAsi;
	  for (set<Alignment*>::iterator oldIt = matches->begin(); oldIt != matches->end(); ++oldIt){ delete (*oldIt); }
	  matches->clear();
	}
	if (matches->size()==0 or newBestScore >= bestScoreSoFar){
	  if (sense == '+'){
	    alignCand->seqReplace(seq, alignCand->seqB());
	    matches->insert(alignCand);
	  } else {
	    matches->insert( alignCand->copyRcSeqA(seq) );
	    //matches->insert( reverseAlignmentSeqA(alignCand, seq) );
	    delete alignCand;
	  }
	} else { delete alignCand; }
      }
    }
    delete localAsi;
    delete [] adequateOffsets;
  }

  delete alMaker;
  delete seqFlip;
}



bool ScoredSeqCollectionBwt::SortWqByLength::operator() (WordQuery* wqA, WordQuery* wqB){
  return wqA->size() < wqB->size();
}


ScoredSeqCollectionBwt::OffsetTracker** ScoredSeqCollectionBwt::getOffsets(ScoredSeq* seq, char sense, long minOverlap,
									   MatchSeqTest* matchTest, AlignmentType alType){

  // this is what I am collecting now so that I can sort the return values
  map<ScoredSeq*,long> hitToIndex;
  long currentIndex = 0;
  OffsetTracker** hitArray = new OffsetTracker*[ _allSeqs.size() ];

  long querySize = seq->size();
  char* seqString = seq->getSeq(sense);  
  // get the size and position of the sub-strings for which
  // matches will be sought
  set<WordQuery*> wordQuerySet;
  if (alType == semiGlobal){
    getSubtextWordQueries(seq, seqString, &wordQuerySet, sense, minOverlap);
  } else if (alType == fullAlignment){
    getSubtextWordQueriesFullAlignment(seq, seqString, &wordQuerySet, sense, minOverlap);
  } else {
    throw AssemblyException::ArgError("SSCBwt:getOffsets got a bad alType");
  }
  // i need to create this array, but i don't need to fill it if there are no good word queries
  long* queryNumArray = new long[querySize];
  if (wordQuerySet.size() > 0){ burrowsWheelerUtilities::makeNumString( seqString, querySize, 0, queryNumArray ); }
  delete [] seqString;
  if (wordQuerySet.size() > 0){

    // here is the block data structure that will be used to keep track of redundant queries
    WordQueryBinOrganizer* wqbo = new WordQueryBinOrganizer(&wordQuerySet);

    // CREATE THE LENGTH-SORTED VERSION OF THIS COLLECTION. SHORTEST ONES FIRST!!!!!
    vector<WordQuery*> sortedWordQueries;
    sortedWordQueries.insert( sortedWordQueries.end(), wordQuerySet.begin(), wordQuerySet.end() );
    SortWqByLength wqSorter; // sort by size, shortest to longest
    sort( sortedWordQueries.begin(), sortedWordQueries.end(), wqSorter );


    for (vector<WordQuery*>::iterator wqIt = sortedWordQueries.begin(); wqIt != sortedWordQueries.end(); ++wqIt){
      WordQuery* wordQuery = *wqIt;

      // make sure that the query is still valid before doing the search
      set<WordQuery*>::iterator foundQuery = wqbo->_querySet.find( wordQuery );
      if (foundQuery != wqbo->_querySet.end()){

	// these will get used a lot
	long stSize = wordQuery->size();
	long startNum = wordQuery->start();

	// remove the query from the collection
	wqbo->_querySet.erase( foundQuery );
	long nb = wqbo->_startToBlock[ startNum ];
	long endEdge = wqbo->_startToBlock[ startNum + stSize ];
	while (nb < endEdge){ wqbo->_blockToQuerySet[nb].erase( wordQuery ); ++nb; }

	// do the search with the sub-query word
	// deal with possible redundant matchLoci if multiple words match
	long* matchArray = burrowsWheelerUtilities::bwtCountAndFind(queryNumArray, startNum, stSize,
								    _bwTransform, _sortedSuffixes, _textSize,
								    _tableC, _occCounts);

	// keep track of the starts AND how many non-overlapping windows have matched
	long numMatches = matchArray[0];
	if (numMatches == 0){
	  long firstBin = wqbo->_startToBlock[ startNum ];
	  long lastBin = wqbo->_startToBlock[ startNum + stSize ] - 1;
	  // copy the first bin so that iteration does not conflict with modification of the set
	  set<WordQuery*> inFirstBin;
	  inFirstBin.insert( wqbo->_blockToQuerySet[firstBin].begin(), wqbo->_blockToQuerySet[firstBin].end() );
	  for (set<WordQuery*>::iterator rwqIt = inFirstBin.begin(); rwqIt != inFirstBin.end(); ++rwqIt){
	    if (firstBin==lastBin or wqbo->_blockToQuerySet[lastBin].count( *rwqIt ) != 0){
	      long nb = wqbo->_startToBlock[ (*rwqIt)->start() ];
	      long endEdge = wqbo->_startToBlock[ (*rwqIt)->start() + (*rwqIt)->size() ];
	      while (nb < endEdge){ wqbo->_blockToQuerySet[nb].erase( *rwqIt ); ++nb; }
	      wqbo->_querySet.erase( *rwqIt );
	    }
	  }
	}
	for (long n = 1; n <= numMatches; ++n){
	  long hitStart = matchArray[n];
	  ScoredSeqLocus* contigLocus = _orderedContigLoci[ getOverlapContigIndex(hitStart, hitStart + stSize - 1) ];
	  ScoredSeq* contig = contigLocus->_seq;
	  if (seq != contig and matchTest->seqIsOk(contig) ){ // don't align to self!!!
	    // determine the offset for the alignment of the query to this seq;
	    // number is the start position of target in seq (i.e. a seq coord)
	    long contigStart = hitStart - contigLocus->_start;

	    // make sure that the hit is on an appropriate region of the target contig
	    if ( wordQuery->acceptableOverlap(contigStart, contig) ){
	      long offset = startNum - contigStart;
	      long contigIndex;
	      map<ScoredSeq*,long>::iterator hitIt = hitToIndex.find( contig );
	      if (hitIt == hitToIndex.end()){
		hitToIndex.insert( pair<ScoredSeq*,long>(contig,currentIndex) );
		contigIndex = currentIndex;
		hitArray[contigIndex] = new OffsetTracker(contig);
		currentIndex++;
	      } else {
		contigIndex = hitIt->second;
	      }
	      hitArray[contigIndex]->addOffset( offset, numMatches );
	    }
	  }
	}
	delete [] matchArray;
      }
    }
    delete wqbo;
    // delete them afterwards, separately
    for (set<WordQuery*>::iterator it = wordQuerySet.begin(); it != wordQuerySet.end(); ++it){
      delete *it;
    }
  }

  // now sort by the seed match counts.  first, figure out how many of each number of counts there are
  map<long,long> hitCountsToCount;
  for (long n = 0; n < currentIndex; ++n){
    long maxOffsetCount = hitArray[n]->maxOffsetCount();
    map<long,long>::iterator it = hitCountsToCount.find( maxOffsetCount );
    if (it == hitCountsToCount.end()){ hitCountsToCount.insert( pair<long,long>(maxOffsetCount,1) ); }
    else { it->second++; }
  }

  // now, figure out the index of the sorted list where each seed count #'s sublist will begin
  // decrement since the indexes (num counts) of the map will be retrieved in increasing order
  map<long,long> hitCountsToStart;
  long currentStart = currentIndex;
  for (map<long,long>::iterator it = hitCountsToCount.begin(); it != hitCountsToCount.end(); ++it){
    currentStart -= it->second;
    hitCountsToStart.insert(pair<long,long>(it->first, currentStart));
  }

  // now, create the new sorted list and fill it
  OffsetTracker** sortedHitWithOffsets = new OffsetTracker*[ currentIndex + 1 ];
  for (long n = 0; n < currentIndex; ++n){
    map<long,long>::iterator it = hitCountsToStart.find(hitArray[n]->maxOffsetCount());
    sortedHitWithOffsets[it->second] = hitArray[n];
    it->second++;
  }
  // the last index is a pair of nulls so that I don't have to return a length
  sortedHitWithOffsets[currentIndex] = NULL;
			     
  delete [] queryNumArray;
  delete [] hitArray;
  return sortedHitWithOffsets;
}


void ScoredSeqCollectionBwt::getSubtextWordQueries(ScoredSeq* querySeq, char* seqSeq, set<WordQuery*>* wordQuerySet, char sense, long minOverlap){
  // local so that filterN's can be done in here but not repeated
  set<WordQuery*> localQuerySet;

  // this is the whole-sequence case
  long blockStartPos = 0;
  long blockEndPos = querySeq->size();
  getSubtextWordQueriesHelper(querySeq, &localQuerySet, sense, blockStartPos, blockEndPos, false);

  // these are the edge cases
  if (_edgeScalingEnabled){
    long fragmentLen = minOverlap;
    while (fragmentLen < querySeq->size()){
      // the front end
      blockStartPos = 0;
      blockEndPos = fragmentLen;
      getSubtextWordQueriesHelper(querySeq, &localQuerySet, sense, blockStartPos, blockEndPos, true);
      // the back end
      blockStartPos = querySeq->size() - fragmentLen;
      blockEndPos = querySeq->size();
      getSubtextWordQueriesHelper(querySeq, &localQuerySet, sense, blockStartPos, blockEndPos, true);
      // scale up the window
      fragmentLen *= 2;
    }
  }

  subtextWordQueriesFilterNs(querySeq, seqSeq, &localQuerySet);
  wordQuerySet->insert(localQuerySet.begin(), localQuerySet.end());
}


void ScoredSeqCollectionBwt::getSubtextWordQueriesFullAlignment(ScoredSeq* querySeq, char* seqSeq, set<WordQuery*>* wordQuerySet,
								char sense, long minTargetLength){
  // local so that filterN's can be done in here but not repeated
  set<WordQuery*> localQuerySet;

  // this is the whole-sequence case; it will allow hits to the entire query to be aligned
  long blockStartPos = 0;
  long blockEndPos = querySeq->size();
  getSubtextWordQueriesHelper(querySeq, &localQuerySet, sense, blockStartPos, blockEndPos, true);

  long fragmentLen = minTargetLength / 2;
  while (fragmentLen < querySeq->size()){
    getSubtextWordQueriesHelper(querySeq, &localQuerySet, sense, fragmentLen);
    // scale up the window
    fragmentLen *= 2;
  }
  subtextWordQueriesFilterNs(querySeq, seqSeq, &localQuerySet);
  wordQuerySet->insert(localQuerySet.begin(), localQuerySet.end());
}


void ScoredSeqCollectionBwt::subtextWordQueriesFilterNs(ScoredSeq* querySeq, char* seqSeq, set<WordQuery*>* wordQuerySet){
  if (wordQuerySet->size() > 0){
    WordQueryBinOrganizer* wqbo = new WordQueryBinOrganizer(wordQuerySet);

    // now do the filtering
    for (long block = 0; block < wqbo->_numBlocks; ++block){
      if (wqbo->_blockToQuerySet[block].size() > 0){
	// scan for N
	long pos = wqbo->_blockToStart[block];
	long blockEnd = wqbo->_blockToStart[block+1];
	while (pos < blockEnd and seqSeq[pos] != 'N'){ ++pos; }
	if (pos != blockEnd){
	  // an N was found, so all overlapping windows can be gotten rid of!
	  for (set<WordQuery*>::iterator wqIt = wqbo->_blockToQuerySet[block].begin(); wqIt != wqbo->_blockToQuerySet[block].end(); ++wqIt){
	    long nb = wqbo->_startToBlock[ (*wqIt)->start() ];
	    // skip this block (and prior blocks) so as to not modify and iterate at the same time
	    if (nb <= block){ nb = block + 1; }
	    long endEdge = wqbo->_startToBlock[ (*wqIt)->start() + (*wqIt)->size() ];
	    while (nb < endEdge){
	      wqbo->_blockToQuerySet[nb].erase( *wqIt );
	      ++nb;
	    }
	    wqbo->_querySet.erase( *wqIt );
	    delete *wqIt;
	  }
	}
      }
    }
    wordQuerySet->clear();
    wordQuerySet->insert(wqbo->_querySet.begin(), wqbo->_querySet.end());
    delete wqbo;
  }
}



ScoredSeqCollectionBwt::WordQueryBinOrganizer::WordQueryBinOrganizer(set<WordQuery*>* wordQueries){
  _querySet.insert(wordQueries->begin(), wordQueries->end());

  // get all the edges
  set<long> edgeSet;
  for (set<WordQuery*>::iterator it = _querySet.begin(); it != _querySet.end(); ++it){
    edgeSet.insert( (*it)->start() );
    edgeSet.insert( (*it)->start() + (*it)->size() );
  }

  // use the edges to construct blocks; they will be ordered in the set
  _blockToStart = new long[ edgeSet.size() ];
  _numBlocks = edgeSet.size() - 1;
  _blockToQuerySet = new set<WordQuery*>[ _numBlocks ];
  long block = 0;
  for (set<long>::iterator it = edgeSet.begin(); it != edgeSet.end(); ++it){
    _blockToStart[block] = *it;
    _startToBlock.insert( pair<long,long>(*it,block) );
    ++block;
  }
  // insert each word query into all of its blocks
  for (set<WordQuery*>::iterator it = _querySet.begin(); it != _querySet.end(); ++it){
    long nb = _startToBlock[ (*it)->start() ];
    long endEdge = _startToBlock[ (*it)->start() + (*it)->size() ];
    while (_blockToStart[nb] < endEdge){
      _blockToQuerySet[nb].insert( *it );
      ++nb;
    }
  }
}
ScoredSeqCollectionBwt::WordQueryBinOrganizer::~WordQueryBinOrganizer(){
  delete [] _blockToStart;
  delete [] _blockToQuerySet;
}



  // num allowed errors:        0  1  2  3  4  5  6  7  8  9  10
int ScoredSeqCollectionBwt::_specialCaseLenDenom[6] =  {1, 2, 3, 3, 4, 5};//, 6, 7, 8, 9, 10};
int ScoredSeqCollectionBwt::_specialCaseStepDenom[6] = {1, 1, 1, 2, 2, 1};//, 1, 1, 1, 1, 1};
int ScoredSeqCollectionBwt::_specialCaseWindowNum[6] = {1, 2, 3, 5, 7, 5};//, 6, 7, 8, 9, 10};

inline long ScoredSeqCollectionBwt::getNumWindowsHelper(long numAllowedMis){
  if (numAllowedMis < 6){ return _specialCaseWindowNum[numAllowedMis]; }
  else {
    // see notebook7 10/26/10 for derivation
    switch(numAllowedMis){
    case 6: return 6;
    case 7: return 6;
    case 8: return 7;
    case 9: return 7;
    case 10: return 8;
    case 11: return 8;
    case 12: return 8;
    case 13: return 9;
    case 14: return 9;
    case 15: return 10;
    case 16: return 10;
    case 17: return 11;
    case 18: return 11;
    case 19: return 11;
    case 20: return 12;
    case 21: return 12;
    case 22: return 13;
    case 23: return 13;
    case 24: return 13;
    case 25: return 14;
    case 26: return 14;
    case 27: return 14;
    case 28: return 15;
    case 29: return 15;
    case 30: return 15;
    case 31: return 16;
    case 32: return 16;
    case 33: return 16;
    case 34: return 17;
    case 35: return 17;
    case 36: return 17;
    case 37: return 18;
    case 38: return 18;
    case 39: return 18;
    case 40: return 19;
    case 41: return 19;
    case 42: return 19;
    case 43: return 20;
    case 44: return 20;
    case 45: return 20;
    case 46: return 21;
    case 47: return 21;
    case 48: return 21;
    case 49: return 21;
    case 50: return 22;
    case 51: return 22;
    case 52: return 22;
    case 53: return 23;
    case 54: return 23;
    case 55: return 23;
    case 56: return 24;
    case 57: return 24;
    case 58: return 24;
    case 59: return 24;
    case 60: return 25;
    case 61: return 25;
    case 62: return 25;
    case 63: return 26;
    case 64: return 26;
    case 65: return 26;
    case 66: return 27;
    case 67: return 27;
    case 68: return 27;
    case 69: return 27;
    case 70: return 28;
    case 71: return 28;
    case 72: return 28;
    case 73: return 29;
    case 74: return 29;
    case 75: return 29;
    case 76: return 29;
    case 77: return 30;
    case 78: return 30;
    case 79: return 30;
    case 80: return 31;
    case 81: return 31;
    case 82: return 31;
    case 83: return 31;
    case 84: return 32;
    case 85: return 32;
    case 86: return 32;
    case 87: return 32;
    case 88: return 33;
    case 89: return 33;
    case 90: return 33;
    case 91: return 34;
    case 92: return 34;
    case 93: return 34;
    case 94: return 34;
    case 95: return 35;
    case 96: return 35;
    case 97: return 35;
    case 98: return 35;
    case 99: return 36;
    case 100: return 36;
    case 101: return 36;
    case 102: return 36;
    case 103: return 37;
    case 104: return 37;
    case 105: return 37;
    case 106: return 38;
    case 107: return 38;
    case 108: return 38;
    case 109: return 38;
    case 110: return 39;
    case 111: return 39;
    case 112: return 39;
    case 113: return 39;
    case 114: return 40;
    case 115: return 40;
    case 116: return 40;
    case 117: return 40;
    case 118: return 41;
    case 119: return 41;
    case 120: return 41;
    case 121: return 41;
    case 122: return 42;
    case 123: return 42;
    case 124: return 42;
    case 125: return 42;
    case 126: return 43;
    case 127: return 43;
    case 128: return 43;
    case 129: return 43;
    case 130: return 44;
    case 131: return 44;
    case 132: return 44;
    case 133: return 45;
    case 134: return 45;
    case 135: return 45;
    case 136: return 45;
    case 137: return 46;
    case 138: return 46;
    case 139: return 46;
    case 140: return 46;
    default: return numAllowedMis / 3;
    }
  }
}




void ScoredSeqCollectionBwt::getSubtextWordQueriesHelper(ScoredSeq* querySeq, set<WordQuery*>* wordQuerySet, char sense,
							 long blockStartPos, long blockEndPos, bool useLimits){
  long seqLen = blockEndPos - blockStartPos;
  bool fiveEdge;
  if (blockStartPos == 0){ fiveEdge = true; }
  else if (blockEndPos == querySeq->size()){ fiveEdge = false; }
  else { throw AssemblyException::ArgError("SSCBwt::getSWQHelper block is at neither the front nor back of the query."); }


  long numAllowedMis = long(float(seqLen) * _maxFractMis);
  // a safety precaution for a rediculous situation
  if (numAllowedMis==seqLen){ numAllowedMis = seqLen - 1; }

  // special cases
  if (numAllowedMis < 6){
    long startDenom = _specialCaseLenDenom[numAllowedMis] * _specialCaseStepDenom[numAllowedMis];
    long subLen = seqLen / _specialCaseLenDenom[numAllowedMis];
    long numWindows = getNumWindowsHelper(numAllowedMis);
    for (long n = 0; n < numWindows; ++n){
      long start = blockStartPos + (n * seqLen / startDenom );
      WordQuery* wc;
      if (! useLimits){ wc = new WordQueryNoEdgeLimit(querySeq, start, subLen, sense); }
      else if (fiveEdge){ wc = new WordQueryFiveEdgeLimit(querySeq, start, subLen, sense, seqLen); }
      else { wc = new WordQueryThreeEdgeLimit(querySeq, start, subLen, sense, seqLen); }
      wordQuerySet->insert(wc);
    }
  } else { // general cases
    long numWindows = getNumWindowsHelper(numAllowedMis);
    long subLen = seqLen / numWindows;
    if (subLen < 2){ subLen = 2; }
    for (long n = 0; n < numWindows; ++n){
      long start = blockStartPos + (n * seqLen / numWindows );
      WordQuery* wc;
      if (! useLimits){ wc = new WordQueryNoEdgeLimit(querySeq, start, subLen, sense); }
      else if (fiveEdge){ wc = new WordQueryFiveEdgeLimit(querySeq, start, subLen, sense, seqLen); }
      else { wc = new WordQueryThreeEdgeLimit(querySeq, start, subLen, sense, seqLen); }
      wordQuerySet->insert(wc);
    }
  }
}



void ScoredSeqCollectionBwt::getSubtextWordQueriesHelper(ScoredSeq* querySeq, set<WordQuery*>* wordQuerySet, char sense,
							 long fragmentLength){

  long numAllowedMis = long(float(fragmentLength) * _maxFractMis);
  // a safety precaution for a rediculous situation
  if (numAllowedMis==fragmentLength){ numAllowedMis = fragmentLength - 1; }
  long querySize = querySeq->size();

  // special cases
  if (numAllowedMis < 6){
    long startDenom = _specialCaseLenDenom[numAllowedMis] * _specialCaseStepDenom[numAllowedMis];
    long subLen = fragmentLength / _specialCaseLenDenom[numAllowedMis];
    long numWindows = getNumWindowsHelper(numAllowedMis) * querySize / fragmentLength;
    for (long n = 0; n < numWindows; ++n){
      long start = n * fragmentLength / startDenom;
      if (start + subLen <= querySize){
	WordQuery* wc = new WordQueryFullTarget(querySeq, start, subLen, sense, _fractId, fragmentLength*2);
	wordQuerySet->insert(wc);
      }
    }
  } else { // general cases
    long numWindows = getNumWindowsHelper(numAllowedMis);
    // adjust for the query length
    long subLen = fragmentLength / numWindows;
    if (subLen < 2){ subLen = 2; }
    numWindows = numWindows * querySize / fragmentLength;
    for (long n = 0; n < numWindows; ++n){
      long start = n * fragmentLength / numWindows;
      if (start + subLen <= querySize){
	WordQuery* wc = new WordQueryFullTarget(querySeq, start, subLen, sense, _fractId, fragmentLength*2);
	wordQuerySet->insert(wc);
      }
    }
  }
}





void ScoredSeqCollectionBwt::bufferSeqs(){
  //OK();
  for ( set<ScoredSeq*>::iterator it = _allSeqs.begin(); it != _allSeqs.end(); ++it ){ (*it)->buffer(); }
  //OK();
}

void ScoredSeqCollectionBwt::unbufferSeqs(){
  //OK();
  for ( set<ScoredSeq*>::iterator it = _allSeqs.begin(); it != _allSeqs.end(); ++it ){ (*it)->unbuffer(); }
  //OK();
}



long ScoredSeqCollectionBwt::size(){
  //OK();
  return _allSeqs.size();
}


// overlaps is the return value, locus is the query
// ASSUMES that the locus only overlaps one contig
long ScoredSeqCollectionBwt::getOverlapContigIndex(long start, long end){
  set<ScoredSeqLocus*> candOverlaps; // all bin members; must be filtered for true overlaps

  // determine the bin to look in
  long startBin = start / _binSize;
  if (startBin < 0){ startBin = 0; }

  // return the legit match's index
  set<ScoredSeqLocus*>::iterator firstLocusIt = _binToSeqs[startBin]->begin();
  while (firstLocusIt != _binToSeqs[startBin]->end() and (! (*firstLocusIt)->overlaps(start,end))){ ++firstLocusIt; }
  if ( firstLocusIt == _binToSeqs[startBin]->end() ){ throw AssemblyException::LogicError("SSCBwt::gOCI can't find match."); }
  return (*firstLocusIt)->_index;
}


ScoredSeqCollectionBwt::ScoredSeqLocus::ScoredSeqLocus(){}
ScoredSeqCollectionBwt::ScoredSeqLocus::ScoredSeqLocus(ScoredSeq* seq, long start, long end, long index) :
  _start(start),
  _end(end),
  _seq(seq),
  _index(index) {
  }
bool ScoredSeqCollectionBwt::ScoredSeqLocus::overlaps(long otherStart, long otherEnd){
  return (! (_start > otherEnd or otherStart > _end) );
}
bool ScoredSeqCollectionBwt::ScoredSeqLocus::contains(long otherStart, long otherEnd){
  return (! (_start > otherStart or otherEnd > _end) );
}



ScoredSeqCollectionBwt::AlignmentMaker::~AlignmentMaker(){}
bool ScoredSeqCollectionBwt::AlignmentMaker::makeAlignmentHelperPlus(ScoredSeq* seqA, long aStart, ScoredSeq* seqB, long bStart,
									   long maxNumMis, long alLength){
  // gaps are irrelevant
  AlignmentScoreMatrix* misCountAsm = new AlignmentScoreMatrix(0,1,1,1);
  long posA = aStart + alLength;
  long posB = bStart + alLength;
  long misCount = 0;
  while ( (maxNumMis >= misCount) and (posA > aStart) ){
    --posA;
    --posB;
    misCount += misCountAsm->match(seqA->nucAtPosPlus(posA), seqB->nucAtPosPlus(posB));
  }
  delete misCountAsm;
  return maxNumMis >= misCount;
}
bool ScoredSeqCollectionBwt::AlignmentMaker::makeAlignmentHelperMinus(ScoredSeq* seqA, long aStart, ScoredSeq* seqB, long bStart,
									    long maxNumMis, long alLength){
  // gaps are irrelevant
  AlignmentScoreMatrix* misCountAsm = new AlignmentScoreMatrix(0,1,1,1);
  long posA = aStart + alLength;
  long posB = bStart + alLength;
  long misCount = 0;
  while ( (maxNumMis >= misCount) and (posA > aStart) ){
    --posA;
    --posB;
    misCount += misCountAsm->match(seqA->nucAtPosPlus(posA), seqB->nucAtPosMinus(posB));
  }
  delete misCountAsm;
  return maxNumMis >= misCount;
}



ScoredSeqCollectionBwt::AlMakerNoEdge::AlMakerNoEdge(){}
ScoredSeqCollectionBwt::AlMakerNoEdge::AlMakerNoEdge(AlignmentScoreMatrix* scoreMatrix, long minScore, float maxFractMis) :
  _scoreMatrix(scoreMatrix),
  _minScore(minScore),
  _maxFractMis(maxFractMis)
{
  if (_scoreMatrix->_match <= _scoreMatrix->_mismatch){
    throw AssemblyException::ArgError("SSCBwt::AlMakerNoEdge needs a lower mismatch score than match score (i.e. mismatch should be negative)");
  }
}
ScoredSeqCollectionBwt::AlMakerNoEdge::~AlMakerNoEdge(){}
Alignment* ScoredSeqCollectionBwt::AlMakerNoEdge::makeAlignment(ScoredSeq* seqA, ScoredSeq* seqB, char sense, long offset){
  // get the starts and alignment length
  long aStart;
  long bStart;
  if (offset < 0){
    aStart = 0;
    bStart = 0 - offset;
  } else {
    aStart = offset;
    bStart = 0;
  }
  long aEnd;
  if (offset + seqB->size() >= seqA->size()){ aEnd = seqA->size(); }
  else { aEnd = offset + seqB->size(); }
  long alLength = aEnd - aStart;

  // i can maybe take out these tests
  if (aStart != 0 and bStart != 0){
    throw AssemblyException::ArgError("in SSCBwt::AlMaker, neither start is zero");
  }
  if ( aStart + alLength != seqA->size() and bStart + alLength != seqB->size() ){
    throw AssemblyException::ArgError("in SSCBwt::AlMaker, neither end is the end of the sequence");
  }

  // the min score must be beaten, and the fractId must also be satisfied
  long maxNumMisByScore = (_minScore - _scoreMatrix->_match * alLength) / (_scoreMatrix->_mismatch - _scoreMatrix->_match );
  long maxNumMisByFid = long(float(alLength) * _maxFractMis);
  long maxNumMis;
  if (maxNumMisByScore < maxNumMisByFid){ maxNumMis = maxNumMisByScore; }
  else { maxNumMis = maxNumMisByFid; }

  bool alIsGood;
  switch (sense){
  case '+': alIsGood = makeAlignmentHelperPlus(seqA, aStart, seqB, bStart, maxNumMis, alLength); break;
  case '-': alIsGood = makeAlignmentHelperMinus(seqA, aStart, seqB, bStart, maxNumMis, alLength); break;
  default: throw AssemblyException::ArgError("SSCBwt:AlMakerNoEdge:makeAl sense char is bad");
  }
  if (alIsGood){ return new AlignmentUngapped(seqA, seqB, sense, offset); }
  else { return new AlignmentNull(seqA, seqB, sense); }
}
long ScoredSeqCollectionBwt::AlMakerNoEdge::getScore(Alignment* al){
  return al->score(_scoreMatrix, false);
}
long ScoredSeqCollectionBwt::AlMakerNoEdge::getMinScore(){ return _minScore; }
void ScoredSeqCollectionBwt::AlMakerNoEdge::setMinScore(long minScore){ _minScore = minScore; }
AlignmentScoreMatrix* ScoredSeqCollectionBwt::AlMakerNoEdge::getScoreMatrix(){ return _scoreMatrix; }



ScoredSeqCollectionBwt::AlMakerPenalizeEdgeA::AlMakerPenalizeEdgeA(){}
ScoredSeqCollectionBwt::AlMakerPenalizeEdgeA::AlMakerPenalizeEdgeA(AlignmentScoreMatrix* scoreMatrix, long minScore, float maxFractMis) :
  _scoreMatrix(scoreMatrix),
  _minScore(minScore),
  _maxFractMis(maxFractMis)
{
  if (_scoreMatrix->_match <= _scoreMatrix->_mismatch){
    throw AssemblyException::ArgError("SSCBwt::AlMakerNoEdge needs a lower mismatch score than match score (i.e. mismatch should be negative)");
  }
}
ScoredSeqCollectionBwt::AlMakerPenalizeEdgeA::~AlMakerPenalizeEdgeA(){}

Alignment* ScoredSeqCollectionBwt::AlMakerPenalizeEdgeA::makeAlignment(ScoredSeq* seqA, ScoredSeq* seqB, char sense, long offset){
  // get the starts and alignment length
  long aStart;
  long bStart;
  if (offset < 0){
    aStart = 0;
    bStart = 0 - offset;
  } else {
    aStart = offset;
    bStart = 0;
  }
  long aEnd;
  if (offset + seqB->size() >= seqA->size()){ aEnd = seqA->size(); }
  else { aEnd = offset + seqB->size(); }
  long alLength = aEnd - aStart;

  // i can maybe take out these tests
  if (aStart != 0 and bStart != 0){
    throw AssemblyException::ArgError("in SSCBwt::AlMaker, neither start is zero");
  }
  if ( aStart + alLength != seqA->size() and bStart + alLength != seqB->size() ){
    throw AssemblyException::ArgError("in SSCBwt::AlMaker, neither end is the end of the sequence");
  }

  // the min score must be beaten, and the fractId must also be satisfied
  long maxNumMisByScore = (_minScore - _scoreMatrix->_match * seqA->size()) / (_scoreMatrix->_mismatch - _scoreMatrix->_match );
  long maxNumMisByFid = long(float(seqA->size()) * _maxFractMis);
  long maxNumMis;
  if (maxNumMisByScore < maxNumMisByFid){ maxNumMis = maxNumMisByScore; }
  else { maxNumMis = maxNumMisByFid; }
  // adjust for the edges
  maxNumMis -= seqA->size() - alLength;

  bool alIsGood;
  switch (sense){
  case '+': alIsGood = makeAlignmentHelperPlus(seqA, aStart, seqB, bStart, maxNumMis, alLength); break;
  case '-': alIsGood = makeAlignmentHelperMinus(seqA, aStart, seqB, bStart, maxNumMis, alLength); break;
  default: throw AssemblyException::ArgError("SSCBwt:AlMakerNoEdge:makeAl sense char is bad");
  }
  if (alIsGood){ return new AlignmentUngapped(seqA, seqB, sense, offset); }
  else { return new AlignmentNull(seqA, seqB, sense); }
}
long ScoredSeqCollectionBwt::AlMakerPenalizeEdgeA::getScore(Alignment* al){
  return al->scoreOverhangA(_scoreMatrix);
}
long ScoredSeqCollectionBwt::AlMakerPenalizeEdgeA::getMinScore(){ return _minScore; }
void ScoredSeqCollectionBwt::AlMakerPenalizeEdgeA::setMinScore(long minScore){ _minScore = minScore; }
AlignmentScoreMatrix* ScoredSeqCollectionBwt::AlMakerPenalizeEdgeA::getScoreMatrix(){ return _scoreMatrix; }




ScoredSeqCollectionBwt::OffsetTracker::OffsetTracker(ScoredSeq* target) : _targetContig(target){
  _maxOffsetCount = 0;
  // this value is meaningless unless the above value is >0
  _maxCountOffset = 0;
}
ScoredSeqCollectionBwt::OffsetTracker::~OffsetTracker(){}
void ScoredSeqCollectionBwt::OffsetTracker::addOffset(long offset, long normFactor){
  float localCount;
  if (normFactor < 1){ throw AssemblyException::ArgError("SSCBwt::OffsetTracker::addOffset, normFactor cannot be less than 1"); }
  float localAddition = float(1) / float(normFactor);
  // now find or create the offset entry
  map<long,float>::iterator offsetEntry = _offsetToCount.find(offset);
  if (offsetEntry == _offsetToCount.end()){
    localCount = localAddition;
    _offsetToCount.insert( pair<long,float>(offset,localCount) );
  } else {
    offsetEntry->second += localAddition;
    localCount = offsetEntry->second;
  }
  if (localCount > _maxOffsetCount){
    _maxOffsetCount = localCount;
    _maxCountOffset = offset;
  }
}
long ScoredSeqCollectionBwt::OffsetTracker::maxOffsetCount(){ return _maxOffsetCount; }
long ScoredSeqCollectionBwt::OffsetTracker::maxCountOffset(){ return _maxCountOffset; }
long* ScoredSeqCollectionBwt::OffsetTracker::getOffsets(){
  // figure out how many offsets there are with each count
  map<float,long> countToNum;
  for (map<long,float>::iterator it = _offsetToCount.begin(); it != _offsetToCount.end(); ++it){
    map<float,long>::iterator foundCtm = countToNum.find( it->second );
    if (foundCtm == countToNum.end()){ countToNum.insert(pair<float,long>(it->second,1)); }
    else { foundCtm->second += 1; }
  }
  // figure out the index in the sorted array at which each offset count will begin
  map<float,long> countToIndex;
  long index = _offsetToCount.size();
  for (map<float,long>::iterator it = countToNum.begin(); it != countToNum.end(); ++it){
    index -= it->second;
    countToIndex.insert(pair<float,long>(it->first,index));
  }
  // fill in the array so it is sorted
  long* offsetArray = new long[_offsetToCount.size() + 1];
  for (map<long,float>::iterator it = _offsetToCount.begin(); it != _offsetToCount.end(); ++it){
    map<float,long>::iterator foundCtm = countToIndex.find( it->second );
    offsetArray[ foundCtm->second ] = it->first;
    foundCtm->second += 1;
  }
  return offsetArray;
}

long ScoredSeqCollectionBwt::OffsetTracker::numOffsets(){ return long(_offsetToCount.size()); }

ScoredSeq* ScoredSeqCollectionBwt::OffsetTracker::targetContig(){ return _targetContig; }


ScoredSeqCollectionBwt::WordQuery::~WordQuery(){}

ScoredSeqCollectionBwt::WordQueryNoEdgeLimit::WordQueryNoEdgeLimit(ScoredSeq* sourceSeq, long start, long size, char sense) :
  _sourceSeq(sourceSeq), _start(start), _size(size), _sense(sense)
{}
ScoredSeqCollectionBwt::WordQueryNoEdgeLimit::~WordQueryNoEdgeLimit(){}
long ScoredSeqCollectionBwt::WordQueryNoEdgeLimit::start(){ return _start; }
long ScoredSeqCollectionBwt::WordQueryNoEdgeLimit::size(){ return _size; }
bool ScoredSeqCollectionBwt::WordQueryNoEdgeLimit::acceptableOverlap(long targetStart, ScoredSeq* target){ return true; }


ScoredSeqCollectionBwt::WordQueryFiveEdgeLimit::WordQueryFiveEdgeLimit(ScoredSeq* sourceSeq, long start, long size, char sense, long edgeLimit) :
  _sourceSeq(sourceSeq), _start(start), _size(size), _sense(sense), 
  _edgeLimit(edgeLimit) {}
ScoredSeqCollectionBwt::WordQueryFiveEdgeLimit::~WordQueryFiveEdgeLimit(){}
long ScoredSeqCollectionBwt::WordQueryFiveEdgeLimit::start(){ return _start; }
long ScoredSeqCollectionBwt::WordQueryFiveEdgeLimit::size(){ return _size; }
bool ScoredSeqCollectionBwt::WordQueryFiveEdgeLimit::acceptableOverlap(long targetStart, ScoredSeq* target){
  long distance = target->size() - targetStart;
  return distance <= _edgeLimit * 2;
}
ScoredSeqCollectionBwt::WordQueryThreeEdgeLimit::WordQueryThreeEdgeLimit(ScoredSeq* sourceSeq, long start, long size, char sense, long edgeLimit) :
  _sourceSeq(sourceSeq), _start(start), _size(size), _sense(sense), 
  _edgeLimit(edgeLimit) {}
ScoredSeqCollectionBwt::WordQueryThreeEdgeLimit::~WordQueryThreeEdgeLimit(){}
long ScoredSeqCollectionBwt::WordQueryThreeEdgeLimit::start(){ return _start; }
long ScoredSeqCollectionBwt::WordQueryThreeEdgeLimit::size(){ return _size; }
bool ScoredSeqCollectionBwt::WordQueryThreeEdgeLimit::acceptableOverlap(long targetStart, ScoredSeq* target){
  long distance = targetStart + _size;
  return distance <= _edgeLimit * 2;
}


ScoredSeqCollectionBwt::WordQueryFullTarget::WordQueryFullTarget(ScoredSeq* sourceSeq, long start, long size, char sense, float fractId, long maxTargetSize) :
  _sourceSeq(sourceSeq), _start(start), _size(size), _sense(sense), _fractId(fractId), _maxTargetSize(maxTargetSize)
{}
ScoredSeqCollectionBwt::WordQueryFullTarget::~WordQueryFullTarget(){}
long ScoredSeqCollectionBwt::WordQueryFullTarget::start(){ return _start; }
long ScoredSeqCollectionBwt::WordQueryFullTarget::size(){ return _size; }
bool ScoredSeqCollectionBwt::WordQueryFullTarget::acceptableOverlap(long targetStart, ScoredSeq* target){
  //the target must be smaller than the source sequence, excpting for the fractId compensation
  if (float(_sourceSeq->size()) < float(target->size()) * _fractId){ return false; }
  else { 
    long targetTrim = long((float(1) - _fractId) * float(target->size()) );
    return (targetStart - targetTrim <= _start) and 
      (_start + target->size() - targetStart <= _sourceSeq->size() + targetTrim);
  }
}


#endif


